//------------------------------------------------------
// module  : Denoise.cc
// auteur  : Mignotte Max
// date    : 
// version : 1.0
// langage : C++
// note    :
//------------------------------------------------------
//  

//------------------------------------------------
// FICHIERS INCLUS -------------------------------
//------------------------------------------------
#include "Denoise.h"

//------------------------------------------------
// DEFINITIONS -----------------------------------
//------------------------------------------------

//------------------------------------------------
// CONSTANTES ------------------------------------
//------------------------------------------------

const float SIGMA_NOISE=30;

const int   METHOD=4;

const float THRESHOLD_UDWT=36;
const int   NBLEVELS=4;

const float STEPLANDWEBER=0.1;
const int   NBITERLANDWEBER=17;

const int ZOOM=1;
const int QUIT=0;
const int SAVE=1;
const int ARRET=1;

//------------------------------------------------
// VARIABLES GLOBALES ----------------------------
//------------------------------------------------
char Name_img[1000];

float SigmaNoise;

int   Method;

float Threshold_Udwt;
int   NbLevels;

float StepLandweber;
int   NbIterLandweber;

int zoom;
int flag_quit;
int flag_save;
int flag_arret;

//----------------------------------------
//--- Windows ----------------------------
//----------------------------------------
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/Xatom.h>
#include <X11/cursorfont.h>

Display   *display;
int	  screen_num;
int 	  depth;
Window	  root;
Visual*	  visual;
GC	  gc;

/************************************************************************/
/* OPEN_DISPLAY()							*/
/************************************************************************/
int open_display()
{
  if ((display=XOpenDisplay(NULL))==NULL)
   { printf("Connection impossible\n");
     return(-1); }

  else
   { screen_num=DefaultScreen(display);
     visual=DefaultVisual(display,screen_num);
     depth=DefaultDepth(display,screen_num);
     root=RootWindow(display,screen_num);
     return 0; }
}

/************************************************************************/
/* FABRIQUE_WINDOW()							*/
/* Cette fonction cre une fenetre X et l'affiche  l'cran.	        */
/************************************************************************/
Window fabrique_window(char *nom_fen,int x,int y,int width,int height,int zoom)
{
  Window                 win;
  XSizeHints      size_hints;
  XWMHints          wm_hints;
  XClassHint     class_hints;
  XTextProperty  windowName, iconName;

  char *name=nom_fen;

  if(zoom<0) { width/=-zoom; height/=-zoom; }
  if(zoom>0) { width*=zoom;  height*=zoom;  }

  win=XCreateSimpleWindow(display,root,x,y,width,height,1,0,255);

  size_hints.flags=PPosition|PSize|PMinSize;
  size_hints.min_width=width;
  size_hints.min_height=height;

  XStringListToTextProperty(&name,1,&windowName);
  XStringListToTextProperty(&name,1,&iconName);
  wm_hints.initial_state=NormalState;
  wm_hints.input=True;
  wm_hints.flags=StateHint|InputHint;
  class_hints.res_name=nom_fen;
  class_hints.res_class=nom_fen;

  XSetWMProperties(display,win,&windowName,&iconName,
                   NULL,0,&size_hints,&wm_hints,&class_hints);

  gc=XCreateGC(display,win,0,NULL);

  XSelectInput(display,win,ExposureMask|KeyPressMask|ButtonPressMask| 
               ButtonReleaseMask|ButtonMotionMask|PointerMotionHintMask| 
               StructureNotifyMask);

  XMapWindow(display,win);
  return(win);
}

/****************************************************************************/
/* CREE_XIMAGE2()							    */
/* Cre une XImage  partir d'un tableau de float                           */
/* L'image peut subir un zoom.						    */
/****************************************************************************/
XImage* cree_Ximage2(float** mat,int z,int length,int width)
{
  int lgth,wdth,lig,col,zoom_col,zoom_lig;
  float somme;
  unsigned char	 pix;
  unsigned char* dat;
  XImage* imageX;

  /*Zoom positiv*/
  /*------------*/
  if (z>0)
  {
   lgth=length*z;
   wdth=width*z;

   dat=(unsigned char*)malloc(lgth*(wdth*4)*sizeof(unsigned char));
   if (dat==NULL)
      { printf("Impossible d'allouer de la memoire.");
        exit(-1); }

  for(lig=0;lig<lgth;lig=lig+z) for(col=0;col<wdth;col=col+z)
   { 
    pix=(unsigned char)mat[lig/z][col/z];
    for(zoom_lig=0;zoom_lig<z;zoom_lig++) for(zoom_col=0;zoom_col<z;zoom_col++)
      { 
       dat[((lig+zoom_lig)*wdth*4)+((4*(col+zoom_col))+0)]=pix;
       dat[((lig+zoom_lig)*wdth*4)+((4*(col+zoom_col))+1)]=pix;
       dat[((lig+zoom_lig)*wdth*4)+((4*(col+zoom_col))+2)]=pix;
       dat[((lig+zoom_lig)*wdth*4)+((4*(col+zoom_col))+3)]=pix; 
       }
    }
  } /*--------------------------------------------------------*/

  /*Zoom negatifv*/
  /*------------*/
  else
  {
   z=-z;
   lgth=(length/z);
   wdth=(width/z);

   dat=(unsigned char*)malloc(lgth*(wdth*4)*sizeof(unsigned char));
   if (dat==NULL)
      { printf("Impossible d'allouer de la memoire.");
        exit(-1); }

  for(lig=0;lig<(lgth*z);lig=lig+z) for(col=0;col<(wdth*z);col=col+z)
   {  
    somme=0.0;
    for(zoom_lig=0;zoom_lig<z;zoom_lig++) for(zoom_col=0;zoom_col<z;zoom_col++)
     somme+=mat[lig+zoom_lig][col+zoom_col];
           
     somme/=(z*z);    
     dat[((lig/z)*wdth*4)+((4*(col/z))+0)]=(unsigned char)somme;
     dat[((lig/z)*wdth*4)+((4*(col/z))+1)]=(unsigned char)somme;
     dat[((lig/z)*wdth*4)+((4*(col/z))+2)]=(unsigned char)somme;
     dat[((lig/z)*wdth*4)+((4*(col/z))+3)]=(unsigned char)somme; 
   }
  } /*--------------------------------------------------------*/

  imageX=XCreateImage(display,visual,depth,ZPixmap,0,(char*)dat,wdth,lgth,16,wdth*4);
  return (imageX);
}

/****************************************************************************/
/* CREE_XIMAGE4()							    */
/* Cre une XImage  partir d'un tableau d'entier                           */
/* L'image peut subir un zoom.						    */
/****************************************************************************/
XImage* cree_Ximage4(int** mat,int z,int length,int width)
{
  int lgth,wdth,lig,col,zoom_col,zoom_lig;
  float somme;
  unsigned char	 pix;
  unsigned char* dat;
  XImage* imageX;

  /*Zoom positif*/
  /*------------*/
  if (z>0)
  {
   lgth=length*z;
   wdth=width*z;

   dat=(unsigned char*)malloc(lgth*(wdth*4)*sizeof(unsigned char));
   if (dat==NULL)
      { printf("Impossible d'allouer de la memoire.");
        exit(-1); }

  for(lig=0;lig<lgth;lig=lig+z) for(col=0;col<wdth;col=col+z)
   { 
    pix=(unsigned char)mat[lig/z][col/z];
    for(zoom_lig=0;zoom_lig<z;zoom_lig++) for(zoom_col=0;zoom_col<z;zoom_col++)
      { 
       dat[((lig+zoom_lig)*wdth*4)+((4*(col+zoom_col))+0)]=pix;
       dat[((lig+zoom_lig)*wdth*4)+((4*(col+zoom_col))+1)]=pix;
       dat[((lig+zoom_lig)*wdth*4)+((4*(col+zoom_col))+2)]=pix;
       dat[((lig+zoom_lig)*wdth*4)+((4*(col+zoom_col))+3)]=pix; 
       }
    }
  } /*--------------------------------------------------------*/

  /*Zoom negatif*/
  /*------------*/
  else
  {
   z=-z;
   lgth=(length/z);
   wdth=(width/z);

   dat=(unsigned char*)malloc(lgth*(wdth*4)*sizeof(unsigned char));
   if (dat==NULL)
      { printf("Impossible d'allouer de la memoire.");
        exit(-1); }

  for(lig=0;lig<(lgth*z);lig=lig+z) for(col=0;col<(wdth*z);col=col+z)
   {  
    somme=0.0;
    for(zoom_lig=0;zoom_lig<z;zoom_lig++) for(zoom_col=0;zoom_col<z;zoom_col++)
     somme+=mat[lig+zoom_lig][col+zoom_col];
           
     somme/=(z*z);    
     dat[((lig/z)*wdth*4)+((4*(col/z))+0)]=(unsigned char)somme;
     dat[((lig/z)*wdth*4)+((4*(col/z))+1)]=(unsigned char)somme;
     dat[((lig/z)*wdth*4)+((4*(col/z))+2)]=(unsigned char)somme;
     dat[((lig/z)*wdth*4)+((4*(col/z))+3)]=(unsigned char)somme; 
   }
  } /*--------------------------------------------------------*/

  imageX=XCreateImage(display,visual,depth,ZPixmap,0,(char*)dat,wdth,lgth,16,wdth*4);
  return (imageX);
}

//----End Windows---------------------------------------------------------
//------------------------------------------------------------------------

//---------------------------------------------------------
// Enregistre les arguments de la ligne de commande      
//---------------------------------------------------------
void read_arguments(int argc,char** argv)
{
 int i;

   //Options
  if (argc<2) 
    { printf("\n Usage Denoise [Img]"); 
      printf("\n       Options are : value indicated by default in ()");
      printf("\n");
      printf("\n Degradation");
      printf("\n -----------");
      printf("\n -n [SigmaNoise]  (> Default:[%.3f])",SIGMA_NOISE); 
      printf("\n");
      printf("\n Method");
      printf("\n ------");
      printf("\n -m ? Method  (> Default[4])");
      printf("\n       [4]  UDWT Deconv. Denoising");
      printf("\n       [20] Denoising UDWT (without deconvolution)");
      printf("\n");
      printf("\n UDWT Parameters");
      printf("\n ---------------"); 
      printf("\n -u ?  UDWTThreshold      (> Default:[%.2f])",THRESHOLD_UDWT);
      printf("\n -x ?  Nomber of levels   (> Default:[%d])",NBLEVELS);
      printf("\n ");
      printf("\n Deconv. Parameters");
      printf("\n ------------------"); 
      printf("\n -g ? Step Gradient Descent Landweber  (> Default:[%.2f])",STEPLANDWEBER);     
      //printf("\n -b ? Nb Iters Landweber               (> Default:[%d])",NBITERLANDWEBER);
      printf("\n --------------");
      printf("\n -r Stopping  (> Default:[%d])",ARRET);
      printf("\n       > [0] ..... End of Iterations"); 
      printf("\n       > [1] ..... Best of MSE Results"); 
      printf("\n       > [2] ..... With Stopping Criterion"); 
      printf("\n -------------");
      printf("\n -z Zoom  (> Default:[%d])",ZOOM);
      printf("\n -q Quit  (> Default:[%d])",QUIT); 
      printf("\n -s Save  (> Default:[%d])",SAVE); 
      printf("\n -------------");
      printf("\n Examples:");
      printf("\n          Denoise lena512.pgm -n 30 -u 36 -q");
      printf("\n          Denoise lena512.pgm -n 25 -u 28 -q");
      printf("\n          Denoise lena512.pgm -n 20 -u 23 -q");
      printf("\n          Denoise lena512.pgm -n 15 -u 17 -q");
      printf("\n          Denoise lena512.pgm -n 10 -u 12 -q");
      printf("\n          ");
      printf("\n          Denoise boat512.pgm -n 30 -u 36 -q");
      printf("\n          Denoise boat512.pgm -n 25 -u 28 -q");
      printf("\n          Denoise boat512.pgm -n 20 -u 23 -q");
      printf("\n          Denoise boat512.pgm -n 15 -u 17 -q");
      printf("\n          Denoise boat512.pgm -n 10 -u 12 -q");
      printf("\n\n\n\n");
      printf("\n\n");
      exit(-1); }

      //Chargement Fichier
      strcpy(Name_img,argv[1]);      

   //Boucle
   for(i=2;i<argc;i++)
     {   
          switch(argv[i][1])
          {
           case 'n': SigmaNoise=atof(argv[++i]);      break;
           case 'm': Method=atoi(argv[++i]);          break; 

           case 'u': Threshold_Udwt=atof(argv[++i]);  break;
           case 'x': NbLevels=atoi(argv[++i]);        break;           
           case 'g': StepLandweber=atof(argv[++i]);   break;   
           //case 'b': NbIterLandweber=atoi(argv[++i]); break; 
          
           case 'r': flag_arret=atoi(argv[++i]);      break;
           case 'z': zoom=atoi(argv[++i]);            break;
           case 's': flag_save=1;                     break; 
           case 'q': flag_quit=1;                     break;
           case '?':                                  break;                
          }    
     } 
 }


//----------------------------------------------------------
//----------------------------------------------------------
// PROGRAMME PRINCIPAL -------------------------------------
//----------------------------------------------------------
//----------------------------------------------------------
int main(int argc,char** argv)
{
 int   flag;
 int length,width;

 //Pour Xwindow
 //------------
 XEvent ev;
 Window win_ppicture;
 Window win_ppicture_noise;
 Window win_ppicture_result;
 XImage *x_ppicture;
 XImage *x_ppicture_noise;
 XImage *x_ppicture_result;
 char nomfen_ppicture[1000];
 char nomfen_ppicture_noise[1000]; 
 char nomfen_ppicture_result[1000];

 //Initialisation
 //--------------
 SigmaNoise=SIGMA_NOISE;

 Method=METHOD;

 Threshold_Udwt=THRESHOLD_UDWT;
 NbLevels=NBLEVELS;

 StepLandweber=STEPLANDWEBER;
 NbIterLandweber=NBITERLANDWEBER;

 zoom=ZOOM;
 flag_save=SAVE;
 flag_quit=QUIT;
 flag_arret=ARRET;
 flag=0;

 //Lis Arguments
 //-------------
 read_arguments(argc,argv);

 //Affiche options
 //---------------
 if (1)
   {
    printf("\n\n Info: ");
    printf("\n -------------------------------");
    printf("\n Ouverture de l'image : <%s>",Name_img);
    printf("\n Sigma Noise :          [%.2f]",SigmaNoise); 
    printf("\n Method                 [%d]",Method);
    printf("\n -----"); 
    printf("\n Threshold_Udwt  :      [%.2f]",Threshold_Udwt);    
    printf("\n Nb Niveaux      :      [%d]",NbLevels);
    printf("\n -----");
    printf("\n StepLandweber   :      [%.2f]",StepLandweber);
    printf("\n NbIterLandweber :      [%d]",NbIterLandweber);
    printf("\n -----");
    printf("\n Flag STOP:              [%d]",flag_arret); 
    printf("\n Flag ZOOM :             [%d]",zoom); 
    printf("\n Flag SAUVE:             [%d]",flag_save); 
    printf("\n Flag QUIT:              [%d]",flag_quit); 
    printf("\n -----------------------------\n");
   }

 
//-------------------------------
//---- Chargement des Images -------------------------------
//-------------------------------


 //Recuperation Images & Info
 //--------------------------
  GetLengthWidth(Name_img,&length,&width);
  float** dataf=fmatrix_allocate_2d(length,width);
  float** datafnoise=fmatrix_allocate_2d(length,width);
  load_image_pgm(Name_img,dataf,length,width);
  load_image_pgm(Name_img,datafnoise,length,width);

  //Ajout Bruit
  if (SigmaNoise) 
  add_gaussian_noise(datafnoise,length,width,SigmaNoise*SigmaNoise);

  //Nb Iterations
  NbIterLandweber=MAX(3,((int)((powf(SigmaNoise,3))/1500)+2));
  printf("  Max Numbers of LandWeber Iterations = [%d]",NbIterLandweber);
  

//-------------------------------
//---- Denoising -------------------------------
//-------------------------------

 //Initialisation Objet
 Denoise* denoi;
 denoi=new Denoise(datafnoise,length,width);

 //Options
 (*denoi).Options_UDWT(Threshold_Udwt,NbLevels);
 (*denoi).Options(dataf,SigmaNoise,Name_img,NbIterLandweber,StepLandweber); 
 (*denoi).FlagSave=flag_save;
 (*denoi).FlagStop=flag_arret;

 //Info
 printf("\n  Info Bruit Simule > MSE = [%.2f]",(*denoi).computeMMSE(datafnoise,dataf,length));

 //Deconv-Denoising
 if (Method==4) (*denoi).NonLinearUdwtDeconvDenoising();
     
 //Denoising Wavelet Traditionnelle
 if (Method==20) (*denoi).Denoising_UDWT();
 

//---------------------------------------------------------------
//---------------- visu sous XWINDOW ----------------------------
//---------------------------------------------------------------
 if (!flag_quit)
 {
 //ouverture session graphique
 if (open_display()<0) printf(" Impossible d'ouvrir une session graphique");

 //commentaire dans les photos affichees
 sprintf(nomfen_ppicture_result,"Image Result : %s",Name_img); 
 sprintf(nomfen_ppicture_noise,"Image Noisy: %s",Name_img); 
 sprintf(nomfen_ppicture,"Image : %s",Name_img); 
 
 //creation des fenetres
 win_ppicture_result=fabrique_window(nomfen_ppicture_result,10,10,width,length,zoom);
 win_ppicture_noise=fabrique_window(nomfen_ppicture_noise,10,10,width,length,zoom);
 win_ppicture=fabrique_window(nomfen_ppicture,10,10,width,length,zoom);
 
 //creation des Ximages
 x_ppicture_result=cree_Ximage2((*denoi).Datafiltered,zoom,length,width);
 x_ppicture_noise=cree_Ximage2(datafnoise,zoom,length,width); 
 x_ppicture=cree_Ximage2(dataf,zoom,length,width);

 printf("\n\n Pour quitter,appuyer sur la barre d'espace");

 //boucle d'evenements
  for(;;)
     {
      XNextEvent(display,&ev);
       switch(ev.type)
        {
         case Expose:     
         XPutImage(display,win_ppicture_result,gc,x_ppicture_result,0,0,0,0,
                   x_ppicture_result->width,x_ppicture_result->height);
         XPutImage(display,win_ppicture_noise,gc,x_ppicture_noise,0,0,0,0,
                   x_ppicture_noise->width,x_ppicture_noise->height);
         XPutImage(display,win_ppicture,gc,x_ppicture,0,0,0,0,
                   x_ppicture->width,x_ppicture->height);
         break;

         case KeyPress:
         XDestroyImage(x_ppicture_result);
         XDestroyImage(x_ppicture_noise);
         XDestroyImage(x_ppicture);
         XFreeGC(display,gc);
         XCloseDisplay(display);
         flag=1;
         break;
         }
   if (flag==1) break;
   }
 } 
       
//--------------- fin de la session graphique --------------     
//----------------------------------------------------------

   //Liberation Memoire
   //------------------
   if (dataf)      free_fmatrix_2d(dataf);
   if (datafnoise) free_fmatrix_2d(datafnoise);
  
   //Retour sans probleme
   //--------------------
   printf("\n   C'est fini... \n"); 
   return 0;
 }
 


