//------------------------------------------------------
// module  : f_Denoise.cc
// auteur  : Mignotte Max
// date    : 
// version : 1.0
// langage : C++
// labo    : DIRO
// note    : 
//------------------------------------------------------
//
//

//------------------------------------------------
// FICHIERS INCLUS -------------------------------
//------------------------------------------------
#include "Denoise.h"

//------------------------------------------------
// FONCTIONS DE LA CLASSE Denoise ----------------
//------------------------------------------------
//----------------------------------------------------------
// Constructeur de la classe Denoise
//----------------------------------------------------------
Denoise::Denoise(float** mat,int lgth,int wdth)
 {
   int i,j;

   //Enregistrement
   length=lgth;
   width=wdth;

   //Allocation memoire matrice
   Data=fmatrix_allocate_2d(length,width);
   Datafiltered=fmatrix_allocate_2d(length,width);  
   DataWithoutNoise=fmatrix_allocate_2d(length,width); 

   DataBest=fmatrix_allocate_2d(length,width); 
   DataStop=fmatrix_allocate_2d(length,width);

   DataTmp1=fmatrix_allocate_2d(length,width);
   DataTmp2=fmatrix_allocate_2d(length,width);
   DataTmp3=fmatrix_allocate_2d(length,width);
   DataTmp4=fmatrix_allocate_2d(length,width);

   //Load
   for(i=0;i<length;i++) for(j=0;j<width;j++) Data[i][j]=mat[i][j];

   //Init
   FlagSave=0;
   Iter_Best=0;
   MSE_Best=10000000;
 }

//----------------------------------------------------------
// Destructeur de la classe Denoise
//----------------------------------------------------------
Denoise::~Denoise()
 { 
  //Liberation memoire matrice 
  if (Data)              free_fmatrix_2d(Data); 
  if (Datafiltered)      free_fmatrix_2d(Datafiltered);
  if (DataWithoutNoise)  free_fmatrix_2d(DataWithoutNoise);

  if (DataBest)          free_fmatrix_2d(DataBest);
  if (DataStop)          free_fmatrix_2d(DataStop);

  if (DataTmp1)          free_fmatrix_2d(DataTmp1);
  if (DataTmp2)          free_fmatrix_2d(DataTmp2);
  if (DataTmp3)          free_fmatrix_2d(DataTmp3); 
  if (DataTmp4)          free_fmatrix_2d(DataTmp4);

  //Delete objets
  if (udwt)  delete udwt;
 }

//-------------------//
// Options ----------//
//-------------------//
//----------------------------------------------------------
// Options UDWT        
//----------------------------------------------------------
void Denoise::Options_UDWT(float thrudwt,int Nbl)
{
 //Enregistrement
  Threshold_Udwt=thrudwt;
  NbLevels=Nbl;

  //Objet
  udwt=new Udwt(length,width);
}

//----------------------------------------------------------
// Options        
//----------------------------------------------------------
void Denoise::Options(float** mat,float Sg,char* Nm,int nb,float stplw)
{
 int i,j;
 int lg;

 //Enregistrement
  SigmaNoise=Sg;
  lg=strlen(Nm);
  strncpy(NameImg,Nm,(lg-4));
  strcpy(NameImg,Nm);
  NbIterLandweber=nb;
  StepLandweber=stplw;
  
 //Enregistrement
 for(i=0;i<length;i++) for(j=0;j<width;j++) DataWithoutNoise[i][j]=mat[i][j];
}


//----------------------------------------------------------
//------ DECONVOLUTIONS/DENOISING --------------------------
//----------------------------------------------------------
//----------------------------------------------------------
// NonLinearWaveDeconvDenoising    
//----------------------------------------------------------     
void Denoise::NonLinearUdwtDeconvDenoising()
{
 int k;
 float NrjMLold,NrjMLcur,StabNrjML;
 float StabMatx;
 float Amplit;
 float Mse;
 float MSE_init;
 int flag=0;

 //Init
 //----
 NrjMLold=0.0;
 NrjMLcur=0.0;
 StabNrjML=0.0;

 //Info
 //----;
 printf("\n\n");
 printf("\n  ----------------------------------------- ");
 printf("\n  Methode -NonLinearWaveletDeconvDenoising- ");
 printf("\n  ----------------------------------------- ");
 printf("\n  Sigma          > [%.2f]",SigmaNoise);
 printf("\n  Threshold_Udwt > [%.2f]",Threshold_Udwt);
 printf("\n  NbLevels       > [%d]",NbLevels);
 printf("\n");

 //----------
 //--Step -1-
 //----------
 printf("\n  > Etape Denoising UDWT Wavelet");
 printf("\n    ----------------------------");
 (*udwt).FilteringUDWT(Data,1.0,Threshold_Udwt,NbLevels);
  MSE_init=computeMMSE(Data,DataWithoutNoise,length);
  printf("\n     MSE >>  [%.5f]",MSE_init); 
 //----------

 //----------
 //--Step -2-
 //----------
 printf("\n\n > LandWeber UDWT Deconvolution");
 printf("\n   ----------------------------");

 float** DataEst=fmatrix_allocate_2d(length,width);
 copy_matrix(DataEst,Data,length,width);

 printf("\n        -> MSE > [%.5f]",computeMMSE(DataEst,DataWithoutNoise,length)); 
 ComputeMoy(DataEst,length,width);
 
 for(k=0;k<NbIterLandweber;k++)
    {
      //> h*x[n] > DataTmp1   
      copy_matrix(DataTmp4,DataEst,length,width);
      (*udwt).FilteringUDWT(DataEst,1.0,Threshold_Udwt,NbLevels);
      copy_matrix(DataTmp1,DataEst,length,width); 
      copy_matrix(DataEst,DataTmp4,length,width); 

      //> x[5]-h*x[n] > DataTmp2
      NrjMLold=NrjMLcur;  
      NrjMLcur=SubstractSquareMatrix(DataTmp2,Data,DataTmp1,length);

      //> hdiese* DataTmp2 > DataTmp2 
      Amplit=AmplitudeSquareMatrix(DataEst,length);
      //printf(" >[A:%.1f]",Amplit); 
      RecalSquareMatrix(DataTmp2,Amplit,length);    
      (*udwt).FilteringUDWT(DataTmp2,1.0,Threshold_Udwt,NbLevels); 
       
      //> x[n] + alpha*[...]
      AddSquareMatrix(DataEst,DataEst,DataTmp2,StepLandweber,length);

      //Contraintes [Max-Min]
      ConstraintMaxMin(DataEst,length);

      //Contraintes Sparsity
      if (!(k%(int)(SigmaNoise/5.0)))
	 { (*udwt).FilteringUDWT(DataEst,1.0,2.5,1); 
           printf(" <U>"); }

      //Mesure
      Mse=computeMMSE(DataEst,DataWithoutNoise,length);
      printf("\n Iter:%d -> [MSE ::%.5f]",k,Mse);
      ComputeMoy(DataEst,length,width);  
      StabNrjML=(fabs(NrjMLcur-NrjMLold)/NrjMLcur);
      printf(" [StabML:%.2f]",StabNrjML*1000); 
      StabMatx=StabMatrix(DataEst,DataTmp4,length);
      printf(" [StabMtx:%.2f]",StabMatx*1000000); 

      //Arret
      if (MSE_Best>Mse) 
	 { MSE_Best=Mse;
           copy_matrix(DataBest,DataEst,length,width); 
           printf("*");
           Iter_Best=k; }

       if ((NrjMLcur<(SigmaNoise/2.50))&&(!flag)&&(k>1))
	 { MSE_Stop=Mse; 
           copy_matrix(DataStop,DataEst,length,width);  
           printf("<---"); 
           flag=1;
           Iter_Stop=k; } 
    }

 //Recup
 copy_matrix(Datafiltered,DataEst,length,width);
 SauveGardeComment(DataBest,length,width,Iter_Best);

 //Info
 printf("\n   UDWT Deconvolution Results");
 printf("\n   --------------------------");
 printf("\n   UDW Deconv. Step:: [MSEinit = %.3f] --> [MSEfinal = %.3f]",MSE_init,MSE_Best);;
}


//----------------------------------------------------------
//------------------ DENOISING -----------------------------
//----------------------------------------------------------
//---------------------------------------------------------
//  Denoising_UDWT     
//----------------------------------------------------------     
void Denoise::Denoising_UDWT()
{
 //Info
 //----;
 printf("\n  ------------------ ");
 printf("\n    UDWT WholeImg    ");
 printf("\n  ---------------- \n");
 printf("\n   >> Sigma   = [%.2f]",SigmaNoise);
 printf("\n");
 printf("\n");
 printf("\n  Threshold_Udwt > [%.2f]",Threshold_Udwt);
 printf("\n  NbLevels       > [%d]",NbLevels);
 printf("\n\n");

 //Init
 udwt=new Udwt(length,width);
 printf("\n");

  //Boucle -------------
  (*udwt).FilteringUDWT(Data,1.0,Threshold_Udwt,NbLevels);
  printf("\n   > MSE >>>>>>> [%.5f]",computeMMSE(Data,DataWithoutNoise,length));    
   
  copy_matrix(Datafiltered,Data,length,width);
  SauveGarde(Datafiltered,length,width);
}

//-------------------//
// MESURE -----------//
//-------------------//
//----------------------------------------------------------
// MMSE              
//----------------------------------------------------------     
float Denoise::computeMMSE(float** mat1,float** mat2,int sz)
{
 int i,j;
 float mmse;

 //Boucle 
 mmse=0.0;
 for(i=0;i<sz;i++) for(j=0;j<sz;j++) mmse+=CARRE(mat2[i][j]-mat1[i][j]);

 mmse/=(CARRE(sz));

 //retour
 return mmse;
}

//-------------------//
// GESTION ----------//
//-------------------//
//----------------------------------------------------------
// SAUVEGARDE          
//----------------------------------------------------------  
void Denoise::SauveGarde(float** mat,int lgth,int wdth)
{
 float mse;
 char Buff[100];
 char BuffName[100];

 //Calcul MSE
 mse=computeMMSE(mat,DataWithoutNoise,lgth);

 //Enregistrement
 sprintf(Buff,"_VarNoise_%.1f__MSE_%.2f",SigmaNoise*SigmaNoise,MSE_Best);
 strcpy(BuffName,NameImg);
 strcat(BuffName,Buff);
 //printf("\n    >[%s]",BuffName);

 //Enregistrement
 if (FlagSave) SaveImagePgm(BuffName,(char*)"",mat,wdth,lgth);
}

//----------------------------------------------------------
// SAUVEGARDE          
//----------------------------------------------------------  
void Denoise::SauveGardeComment(float** mat,int lgth,int wdth,int k)
{
 float mse;
 char Buff[100];
 char BuffName[100];

 //Info
 printf(" [%d]",k);

 //Calcul MSE
 mse=computeMMSE(mat,DataWithoutNoise,lgth);

  //Enregistrement
 //printf("\n\n\n    > NameImg  > [%s]",NameImg);
 printf("\n");
 sprintf(Buff,"_VarN%.0f",SigmaNoise*SigmaNoise);
 strcpy(BuffName,NameImg);
 strcat(BuffName,Buff);
 sprintf(Buff,"_MSE%.2f",mse);
 strcat(BuffName,Buff);
 //sprintf(Buff,"_U%.1f",Threshold_Udwt);
 //strcat(BuffName,Buff);
 //sprintf(Buff,"_L%d",NbLevels);
 //strcat(BuffName,Buff);
 //sprintf(Buff,"_it%d",k);
 //printf("\n    >[%s]",BuffName);

 //Enregistrement
 if (FlagSave) SaveImagePgm(BuffName,(char*)"",mat,wdth,lgth);
}

