//---------------------------------------------------
// module  : f_Udwt.cc                         
// auteur  : Max Mignotte                           
// date    :                                              
// langage : C++                                      
// labo    :                                  
//---------------------------------------------------

//---------------------------------------------------
// FICHIERS INCLUS ----------------------------------
//---------------------------------------------------
#include "Udwt.h"

//----------------------------------------------------
// CONSTRUCTEUR ET DESTRUCTEUR DE LA CLASSE Udwt -----
//----------------------------------------------------
//----------------------------------------------------------
// constructeur de la classe Udwt
//----------------------------------------------------------
Udwt::Udwt(int Length,int Width) 
 { 
  //--init. Longeur/Largeur--
  width=Width; length=Length;
  size=width;
  nlev_udwt=(int)(log((float)width)/log(2.0));
  //printf("\n  Objet Udwt");
  //printf("\n     >Nb level UDWT=%d",nlev_udwt);


  //--Width=Length?--
  if (width!=length)
    { printf("\n Length et Width differents");
      exit(-1); }

  //--init matrice --
  mat=fmatrix_allocate_2d(length,width);
  mat_rec=fmatrix_allocate_2d(length,width);

  //--Init Vecteur--
  vect=fmatrix_allocate_1d(width);
  vect_tmp=fmatrix_allocate_1d(width);

  //Init matrice UDWT 
  mat_udwt=fmatrix_allocate_2d(length*nlev_udwt,width*nlev_udwt);
  mat_udwt_rec=fmatrix_allocate_2d(length*nlev_udwt,width*nlev_udwt);
  mat_udwt_dec=fmatrix_allocate_2d(length*nlev_udwt,width*nlev_udwt);

  vect_udwt=fmatrix_allocate_1d(width*nlev_udwt);
  vect_udwt_dec=fmatrix_allocate_1d(width*nlev_udwt); 
  vect_udwt_rec=fmatrix_allocate_1d(width*nlev_udwt);
  vect_udwt_tmp=fmatrix_allocate_1d(width*nlev_udwt);
  vect_iudwt=fmatrix_allocate_1d(width*nlev_udwt);
  vect_iudwt_tmp=fmatrix_allocate_1d(width*nlev_udwt);
  vect_iudwt_tmp2=fmatrix_allocate_1d(width*nlev_udwt); 
  tabvect_iudwt=fmatrix_allocate_2d(2,width*nlev_udwt); 

  //--Flag------
  FlagComment=0;
 }

//----------------------------------------------------------
// destructeur de la classe Udwt
//----------------------------------------------------------
Udwt::~Udwt()
 { 
   //Liberation memoire matrice--
   free_fmatrix_2d(mat); 
   free_fmatrix_2d(mat_rec);

   //Liberation memoire vecteur--
   free_fmatrix_1d(vect); 
   free_fmatrix_1d(vect_tmp);

   //UDWT
   free_fmatrix_2d(mat_udwt);
   free_fmatrix_2d(mat_udwt_rec);
   free_fmatrix_2d(mat_udwt_dec);
 
   free_fmatrix_1d(vect_udwt);
   free_fmatrix_1d(vect_udwt_dec);
   free_fmatrix_1d(vect_udwt_rec);
   free_fmatrix_1d(vect_udwt_tmp);
   free_fmatrix_1d(vect_iudwt);
   free_fmatrix_1d(vect_iudwt_tmp); 
   free_fmatrix_1d(vect_iudwt_tmp2); 
   free_fmatrix_2d(tabvect_iudwt); 
 }


//------------------------------//
//------------UDWT--------------//
//------------------------------//
//----------------------------------------------------------
// UDWT_2D
//----------------------------------------------------------
void Udwt::UDWT_2D(int nlev)
{
 int i,j,k,l;
 int sz;

  //Presentation
  if (FlagComment)
     {
      printf("\n\n\n");
      printf("\n -------------------------");
      printf("\n Decomposition Image UDWT ");
      printf("\n -------------------------");
      printf("\n");
     }

  //Initialisation
  NbLevels=nlev;
  sz=size;
  if (FlagComment) printf("\n Decomposition UDWT en %d Niveaux \n",NbLevels);

  //BOUCLE
  for(k=0;k<NbLevels;k++)
     {
      if (FlagComment) printf("[size:%d]",sz);
      //lignes
      for(i=0;i<sz;i++)   UDWT_1D_(mat_udwt_dec[i],k);
      sz+=size;

      //colonnes
      for(j=0;j<sz;j++) 
         {
          for(l=0;l<sz;l++)       vect_udwt_dec[l]=mat_udwt_dec[l][j];
          UDWT_1D_(vect_udwt_dec,k);
          for(l=0;l<sz;l++)       mat_udwt_dec[l][j]=vect_udwt_dec[l];
         }
     }
}

//----------------------------------------------------------
// IUDWT_2D
//----------------------------------------------------------
void Udwt::IUDWT_2D()
 {
  int i,j,k,l;
  int sz;

  //Presentation
  if (FlagComment)
     {
      printf("\n\n\n");
      printf("\n ---------------------------");
      printf("\n Reconstruction Image IUDWT ");
      printf("\n ---------------------------");
      printf("\n");
     }

  //Initialisation
  if (FlagComment) printf("\n Decomposition inverse en %d Niveaux",NbLevels);

  for(i=0;i<(size*nlev_udwt);i++) for(j=0;j<(size*nlev_udwt);j++) 
  mat_udwt_rec[i][j]=mat_udwt_dec[i][j];

  sz=(int)(size*(NbLevels+1));
  if (FlagComment) printf("\n Size:%d \n",sz);
  
  //BOUCLE
  for(k=(NbLevels-1);k>=0;k--)
     { 
       if (FlagComment) printf("[size:%d]",sz);
      //colonnes
      for(j=0;j<sz;j++) 
         {
          for(l=0;l<sz;l++)       vect_udwt_rec[l]=mat_udwt_rec[l][j];
          IUDWT_1D_(vect_udwt_rec,k);
          for(l=0;l<sz;l++)       mat_udwt_rec[l][j]=vect_udwt_rec[l];
         }

      //lignes
      for(i=0;i<sz;i++) IUDWT_1D_(mat_udwt_rec[i],k);
 
      sz-=size;
     }

 //Copy
 for(i=0;i<length;i++)  for(j=0;j<width;j++) mat_rec[i][j]=mat_udwt_rec[i][j]; 

 //Debug
 //for(i=0;i<length;i++)  for(j=0;j<width;j++)
 //  if  ((mat_rec[i][j]<0)||(mat_rec[i][j]>255)) printf("[%d-%d:%.2f]",i,j,mat_rec[i][j]);
 }


//----------------------------------------------------------
// UDWT_1D_
//----------------------------------------------------------
void Udwt::UDWT_1D_(float* vct,int Nlev)
{
 int k;
 int szv;

 //Initialisation
 szv=(int)(size/pow(2,Nlev));   

 //Copiage Old Difference Vector
 Copy(&vct[size],&vect_udwt[2*size],size*Nlev); 

 //Boucle
 for(k=0;k<pow(2,Nlev);k++)
   { 
    Copy(&vct[k*szv],vect_udwt_tmp,szv);   
    Haar1D_(vect_udwt_tmp,szv);      

    Copy(&vect_udwt_tmp[0],&vect_udwt[k*szv],(szv/2));
    Copy(&vect_udwt_tmp[szv/2],&vect_udwt[size+(k*szv)],(szv/2)); 

    Copy(&vct[k*szv],vect_udwt_tmp,szv);
    Decal(vect_udwt_tmp,szv); 
    Haar1D_(vect_udwt_tmp,szv); 

    Copy(&vect_udwt_tmp[0],&vect_udwt[(k*szv)+(szv/2)],(szv/2)); 
    Copy(&vect_udwt_tmp[szv/2],&vect_udwt[size+(k*szv)+(szv/2)],(szv/2));
   }

 //Copy
 Copy(vect_udwt,vct,size*(Nlev+2)); 
}

//----------------------------------------------------------
// IUDWT_1D_
//----------------------------------------------------------
void Udwt::IUDWT_1D_(float* vct,int Nlev)
{
 int k;
 int szv;

  //Init
  szv=(int)(size/pow(2,Nlev)); 

  //Copiage Old Diff-Vector 
  Copy(&vct[2*size],&vect_iudwt[size],size*Nlev);  

      //Boucle
      for(k=0;k<pow(2,Nlev);k++)
         { 
          Copy(&vct[k*szv],&vect_iudwt_tmp[0],szv/2);
          Copy(&vct[size+(k*szv)],&vect_iudwt_tmp[szv/2],szv/2); 

          HaarInv1D_(vect_iudwt_tmp,szv); 
          Copy(vect_iudwt_tmp,tabvect_iudwt[0],szv); 

          Copy(&vct[(k*szv)+(szv/2)],&vect_iudwt_tmp[0],szv/2); 
          Copy(&vct[size+(k*szv)+(szv/2)],&vect_iudwt_tmp[szv/2],szv/2);

          HaarInv1D_(vect_iudwt_tmp,szv); 
          Recal(vect_iudwt_tmp,szv); 
          Copy(vect_iudwt_tmp,tabvect_iudwt[1],szv); 

          Average(vect_iudwt_tmp2,tabvect_iudwt[0],tabvect_iudwt[1],szv);
          Copy(vect_iudwt_tmp2,&vect_iudwt[k*szv],szv);
         }

  //Copy
  Copy(vect_iudwt,vct,size*(Nlev+1)); 
}


//---------------//
//-- DENOISING --//
//---------------//
//-----------------------------------------------------------------
// 
//  ((w2)-3Tau*Sigma2Bruit)+
//  ------------------------          [ (x)+= max{x,0} ]
//            w   
//
//------------------------------------------------------------------
void Udwt::FilteringUDWT(float** mt,float tau,float sig,int nlev)
{
 int i,j;
 float tmp;
 float signal,blur; //Stat

 //Recup Parametre
 Tau=tau;
 Sigma2Noise=sig*sig;
 NbLevels=nlev; 

 //Initialisation
 if (FlagComment) printf("\n Size:%d",size);
 signal=0.0; blur=0.0;  //Stat

 //Copie Matrice
 for(i=0;i<length;i++) for(j=0;j<width;j++) mat_udwt_dec[i][j]=mt[i][j]; 

 //Decomposition
 UDWT_2D(nlev);

 //Seuillage
 for(i=0;i<(size*nlev_udwt);i++) for(j=0;j<(size*nlev_udwt);j++)
    if (!((i<size)&&(j<size)))
    {
     signal+=CARRE(mat_udwt_dec[i][j]);  //Stat
     tmp=CARRE(mat_udwt_dec[i][j])-3*Tau*Sigma2Noise;

     if (tmp<0) tmp=0.0;
     if (mat_udwt_dec[i][j]!=0.0) mat_udwt_dec[i][j]=(tmp/mat_udwt_dec[i][j]);
     blur+=CARRE(mat_udwt_dec[i][j]);  //Stat
    }

 //Reconstruction
 IUDWT_2D();

 //Copy
 for(i=0;i<length;i++) for(j=0;j<width;j++) mt[i][j]=mat_udwt_rec[i][j];
}

//--------//
//-UTILES-//
//--------//
//----------------------------------------------------------
// Haar1d
//----------------------------------------------------------
void Udwt::Haar1D_(float* vct,int nb)
{
 int k,middle;
  
  //Init & Recup
  middle=(nb/2);
  for(k=0;k<nb;k++) vect_tmp[k]=vct[k];
  
  //Remplissage
  for(k=0;k<middle;k++)
     { vct[k]=((vect_tmp[2*k]+vect_tmp[(2*k)+1])/sqrt(2.0));
       vct[k+middle]=((vect_tmp[2*k]-vect_tmp[(2*k)+1])/sqrt(2.0)); }
}

//----------------------------------------------------------
// HaarInv1d
//----------------------------------------------------------
void Udwt::HaarInv1D_(float* vct,int nb)
{
 int k,middle;
  
  //Init & Recup
  middle=(nb/2);
  for(k=0;k<nb;k++) vect_tmp[k]=vct[k];
  
  //Remplissage
  for(k=0;k<middle;k++)
     { vct[2*k]=((vect_tmp[k]+vect_tmp[k+middle])/sqrt(2));
       vct[(2*k)+1]=((vect_tmp[k]-vect_tmp[k+middle])/sqrt(2));}
}

//----------------------------------------------------------
// Decal Vecteur
//----------------------------------------------------------
void Udwt::Decal(float* vct,int dim)
{
 int k;
 float tmp;

 //Recup
 tmp=vct[(dim-1)];

 //Boucle
 for(k=(dim-2);k>=0;k--) vct[k+1]=vct[k];

 //last
 vct[0]=tmp;
}

//----------------------------------------------------------
// Recal Vecteur
//----------------------------------------------------------
void Udwt::Recal(float* vct,int dim)
{
 int k;
 float tmp;

 //Recup
 tmp=vct[0];

 //Boucle
 for(k=0;k<(dim-1);k++) vct[k]=vct[k+1];

 //last
 vct[(dim-1)]=tmp;
}

//----------------------------------------------------------
// Copy
//----------------------------------------------------------
void Udwt::Copy(float* vctin,float* vctout,int nb)
{
 int k;

 //Boucle
 for(k=0;k<nb;k++) vctout[k]=vctin[k];
}

//----------------------------------------------------------
// Average
//----------------------------------------------------------
void Udwt::Average(float* vct,float* vct1,float* vct2,int nb)
{
 int k;

 //Average
 for(k=0;k<nb;k++) vct[k]=((vct1[k]+vct2[k])/2.0);
}


 




